#!/usr/bin/env python3

# AVR / Arduino dynamic memory log analyis script.
#
# Copyright (c) 2014 Matthijs Kooijman <matthijs@stdin.nl>
#
# This code is licensed under the MIT license:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This script is meant to analyze the output of the corresponding
# Arduino-side malloc, realloc and free logging code. It mostly checks
# for double free problems, and also gives a list of allocations not
# freed at the end.
#
# To run it, feed the log output into stdin, the analysis result will
# appear on stdout.

import io
import re
import sys
import collections

allocations = {}

Allocation = collections.namedtuple('Allocation', ['line', 'lineno', 'size'])

infile = io.TextIOWrapper(sys.stdin.buffer, encoding='latin-1')

for n, line in enumerate(infile.readlines()):
    # Make lines 1-based instead of 0-based
    n += 1

    match = re.match(r"^malloc\(([0-9]+)\) = (0x[0-9A-F]+)$", line)
    if match:
        allocations[match.group(2)] = Allocation(line = line, lineno = n, size = int(match.group(1)))
        continue

    match = re.match(r"^realloc\((0x[0-9A-F]+), ([0-9]+)\) = (0x[0-9A-F]+)$", line)
    if match:
        if match.group(1) != '0x0':
            try:
                allocations.pop(match.group(1))
            except KeyError:
                sys.stdout.write("{:3d}: {}".format(n, line))
                sys.stdout.write(">> {} not allocated\n".format(match.group(1)))
                continue

        allocations[match.group(3)] = Allocation(line = line, lineno = n, size = int(match.group(2)))
        continue

    match = re.match(r"^free\((0x[0-9A-F]+)\)$", line)
    if match:
        if (match.group(1) == '0x0'):
            continue
        try:
            allocations.pop(match.group(1))
        except KeyError:
            sys.stdout.write("{:3d}: {}".format(n, line))
            sys.stdout.write(">> {} not allocated\n".format(match.group(1)))
            continue

        continue

for addr, alloc in allocations.items():
    sys.stdout.write("{:3d}: {}".format(alloc.lineno, alloc.line))
    sys.stdout.write(">> {} never freed\n".format(addr))
